import time
import numpy
import matplotlib.pyplot as plt
import keysightSD1
import keysight_hvi as kthvi

"""
Pathwave Test Sync Executive - Real-Time Pulsed Characterization of a Device-Under-Test

Copyright Keysight Technologies 2022

In this programming example an M320xA AWG and an M3102A digitizer are used to run a physics experiment typical of quantum applications. A pool of different 
waveforms is loaded to the AWG RAM. The digitizer can use the register sharing functionality to select real-time the waveform to be played by the AWG at 
each iteration of the experiment steps. The selected waveform is used by AWG CH1 and CH2 to play I-Q modulated pulses and re-play them after a variable 
delay. In the same iteration AWG CH3 and CH4 play I-Q readout pulses after a variable readout delay. The readout pulse length is increased after each 
iteration. The experiment can be repeated for a user-defined number of loops, allowing the user to choose the delay between each loop, delay necessary 
for example to let the Device-Under-Test (DUT) return to its equilibrium state.

The showcased HVI API functionalities are:
* Off-shelf inter-instrument synchronization capabilities
* HVI Synchronized Multi-Sequence Block to synchronously parallel-execute sequences in multiple instruments
* HVI Synchronized While Statement to run synchronized loops over multiple instruments
* Combination of nested Sync While and Sync Multi-Sequence Blocks
* HVI Local Flow Control: Wait Time Statement, IF Statement
* HVI native and instrument-specific instructions
* Sync Register Sharing functionality to real-time share register information through different instruments

The Python code that follows:
1. Opens an M320xA AWG and an M3102A digitizer (real hardware or simulation mode) 
2. Defines HVI resources (engines, chassis, chassis interconnections, trigger, action, events, HVI triggering resources, etc)
3. Programs HVI sequences by adding the necessary statements and instructions in each AWG sequence
4. Compiles HVI, loads it to HW and executes it

For further info please check the programming example titled "Pathwave Test Sync Executive - Real-Time Pulsed Characterization of a Device-Under-Test"
"""
######################################
# Application Configuration 
######################################

class ApplicationConfig:
    " Defines module descriptors, configuration options and names of HVI engines, actions, triggers"
    def __init__(self):
        """
        Define PXI instruments options
        """
        # Define options to open the instruments. 
        # Complete option list can be found in the SD1 3.x User Guide
        self.options = 'channelNumbering=keysight'
        # Simulated HW Option
        self.hardware_simulated = True
        """ 
        Define names of HVI engines, actions, registers
        """
        # HVI engine names to be used in this application 
        self.awg_engine_name = "AWG Engine"
        self.dig_engine_name = "Digitizer Engine"
        # HVI action names to be used by each HVI engine
        self.awg_trigger_action_name = "AWG_Trigger"
        self.daq_trigger_action_name = "DAQ_Trigger"
        # HVI register names to be used within the scope of each HVI engine
        self.steps_name = "Steps"
        self.loops_name = "Loops"
        self.wfm_id_name = "Waveform ID"
        self.tau_name = "Tau"
        self.pulse_length_name = "Pulse Length"
        self.acq_delay_name = "Acquisition Delay"
        self.awg_counter_name = "AWG Counter"
        self.dig_counter_name = "Digitizer Counter"
        self.hvi_done_name = "HVI Done"
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information
        self.digitizer_descriptor = ModuleDescriptor('M3102A', 1, 6, self.options, self.dig_engine_name)
        self.awg_descriptor = ModuleDescriptor('M3202A', 1, 4, self.options, self.awg_engine_name)
        self.module_descriptors = [self.digitizer_descriptor, self.awg_descriptor]
        
        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [1]
        
        # Multi-chassis setup
        # Define the System Sync Modules included in your system.
        # For example: 
        # self.system_sync_module_descriptors = [
        #    SystemSyncModuleDescriptor('PXI0::CHASSIS2::SLOT10::INDEX0::INSTR', self.ssm_options)]
        self.ssm_options = ''
        self.ssm_simulation_options = 'Simulate=True,DriverSetup=Model=M9033A'
        self.system_sync_modules_descriptors = []
        # For each SSM define which SSM is connected to its downstream connectors.
        # For example: 
        # self.ssm_connections = [
        #      SystemSyncModuleConnection(ssm1_chassis=1, ssm1_downstream_connector_number=0, ssm2_chassis=2)]
        self.ssm_connections = []
        
        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        # NOTE: In a multi-chassis setup ALL the PXI lines listed below need to be shared among each M9031 board pair by means of SMB cable connections
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER0,
            kthvi.TriggerResourceId.PXI_TRIGGER1,
            kthvi.TriggerResourceId.PXI_TRIGGER2, 
            kthvi.TriggerResourceId.PXI_TRIGGER3, 
            kthvi.TriggerResourceId.PXI_TRIGGER4, 
            kthvi.TriggerResourceId.PXI_TRIGGER5, 
            kthvi.TriggerResourceId.PXI_TRIGGER6, 
            kthvi.TriggerResourceId.PXI_TRIGGER7]
        """
        Defines the experiment parameters
        """ 
        self.num_wfms = 1 # Number of waveforms to be selected from the AWG RAM
        self.T2_flag = 0 # User can choose to run a T1 or T2 experiment
        self.initial_tau = 10 # x10[ns] # The initial time delay between the control and readout pulse, in ns
        self.tau_step = 20 # x10[ns] # Time that is incrementally added to delay between the control and readout pulse, in ns
        self.ro_delay = 150 # [ns] # Delay in ns that is applied after the last control pulse, but before the readout pulse
        self.step_delay = 10 # [ns] # Time to wait between each experiment step
        self.loop_delay = 20 # [ns] # Time to wait between each experiment loop
        self.initial_acq_delay = 250 # x10[ns] # Delay before starting to capture waveforms with digitizer
        self.acquisition_window = 2000 # [ns] time window to be acquired by DAQ channel each time a DAQ trigger is sent out
        self.carrier_frequency = 100e6 # [Hz] frequency of the IF carrier modulating the I-Q pulses at the AWG output
        self.initial_pulse_length = 20 # x10[ns] # Initial readout pulse length
        self.delta_length = 10 # x10[ns] # Duration increment of the readout pulse length at each step
        self.num_steps = 5 # Number of iterations to increase tau by tau_step
        self.num_loops = 1 # Number of experiments to execute
        """
        AWG parameters
        """
        self.all_ch_mask = 0xF # binary mask defining which channels to use
        # AWG settings for all channels
        self.sync_mode = keysightSD1.SD_SyncModes.SYNC_NONE
        self.queue_mode = keysightSD1.SD_QueueMode.ONE_SHOT
        self.awg_mode = keysightSD1.SD_Waveshapes.AOU_SINUSOIDAL
        self.start_delay = 0 # x10 [ns]
        self.awg_prescaler = 0
        self.wfm_cycles = 2 # number of pulsed wfms for the T2 experiment
        self.amplitude = 1 # [V]
        self.offset = 0 # [V]
        # Trigger settings
        self.awg_trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG_CYCLE
        # Latency values for M3202A AWGqueueWfm() [ns]
        # Latencies depend on AWG FPGA FW. Please check the SD1 3.x User Guide for detailed info
        self.queue_wfm_latency = 100 # [ns] Minimum start delay necessary to execute an AWGqueueWfm() instruction
        self.awg_trigger_latency = 2300 # [ns] Minimum latency necessary between an AWGqueueWfm() instruction and an AWGtrigger action.
        # Readout pulse parameters
        self.rorise_id = 1000 # wfm ID for the rising edge of the readout pulse
        self.rofall_id = 1001 # wfm ID for the falling edge of the readout pulse
        """
        Digitizer parameters 
        """
        self.sampling_time = 2 # [ns] 1/sample_rate, sample_rate = 500 MSa/s for Digitizer M3102A 
        self.dig_prescaler = 0 # Prescaler values are explained in M3xxxA User Guide
        self.fullscale = 2 # [V] enter x Volts to set the full scale to [-x, x] Volts
        self.acquisition_points_per_cycle = int(self.acquisition_window / self.sampling_time) # [Sa]
        self.num_cycles = self.num_steps*self.num_loops # insert -1 for infinite cycles
        self.acquisition_points = self.acquisition_points_per_cycle*self.num_cycles
        self.acquisition_delay = 0 # x2[ns]
        self.dig_trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG
        self.dig_mask = self.all_ch_mask

######################################
# Main Program
######################################

def run_experiment():
    """ 
    Runs DUT characterization experiment
    """
    # Load configuration and parameters
    config = ApplicationConfig()
    # Checks flags defining if the code executes on HW/Simulation, SD1
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    tau_final = 10*(config.initial_tau + config.tau_step*(config.num_steps-1))
    pulse_length_final = 10*(config.initial_pulse_length + config.delta_length*(config.num_steps-1))
    print()
    print("----------------------")
    print("Experiment Parameters")
    print("----------------------")
    print("Number of HVI loops: {}".format(config.num_loops))
    print("Number of HVI steps: {}".format(config.num_steps))
    print("Carrier frequency: {} MHz".format(int(config.carrier_frequency/1e6)))
    print("Number of different AWG waveforms: {}".format(config.num_wfms))
    print("Digitizer configured to acquire {} points in {} cycles".format(config.acquisition_points_per_cycle, config.num_cycles))
    print("Total dig. acquisition_points = points_per_cycle*num_cycles = {}".format(config.acquisition_points))
    print("Readout delay: {} ns".format(config.ro_delay))
    print("Tau delay set to sweep from {} ns to {} ns in {} steps of {} ns".format(10*config.initial_tau, tau_final, config.num_steps-1, 10*config.tau_step))
    print("Readout pulse duration set to sweep from {} ns to {} ns in {} steps of {} ns".format(10*config.initial_pulse_length, pulse_length_final, config.num_steps-1, 10*config.delta_length))
    print()
    
    ###################################################
    # Open PXI modules, configure AWG and Digitizer
    ###################################################
    # Open instruments objects
    # Returns a dictionary of module objects whose keys are the HVI engine names
    module_dict = open_modules(config)
    
    # Configure Digitizer
    configure_digitizer(config, module_dict[config.dig_engine_name].instrument)
    # Configure AWG
    configure_awg(config, module_dict[config.awg_engine_name].instrument)

    ######################################
    # System definition
    ######################################
    # Create system definition object
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform, add HVI resources
    define_hvi_resources(my_system, module_dict, config)
    # Output to console the details of the system definition
    output_sys_def(my_system, config.module_descriptors)
    
    ######################################
    # Program HVI sequences
    ######################################
    print("-----------------------")
    print("Program HVI Sequences")
    print("-----------------------")
    print("Initializing the defined system...")
    # Create sequencer object
    sequencer = kthvi.Sequencer("MySequencer", my_system)
    print("Programming the HVI sequences...")
    # Program the HVI sequence
    program_dut_experiment(sequencer, module_dict, config)
    # Export the programmed sequence to text
    export_hvi_seqeunces(sequencer)
    
    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    print("-------------------")
    print("Execute HVI")
    print("-------------------")
    print("Compiling HVI...")
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()
    print("HVI Loaded to HW")

    # Execute HVI in non-blocking mode 
    # This mode allows SW execution to interact with HVI execution
    hvi.run(hvi.no_wait)
    print("HVI Running...")

    # User-controlled HVI execution
    if config.hardware_simulated:
        print("Simulation completed successfully")
    else:
        interact_with_hvi(hvi, module_dict, config)

    # Release HW resources once HVI execution is completed
    hvi.release_hw()
    print("Releasing HW...")
    
    # Close all modules at the end of the execution
    for engine_name in module_dict:
        module_dict[engine_name].instrument.close()
    print("PXI modules closed\n")

########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
########################################################

def define_hvi_resources(sys_def, module_dict, config):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)

    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, module_dict)
    
    # Define list of actions to be executed
    define_hvi_actions(sys_def, module_dict, config)


def define_hw_platform(sys_def, config):
    """
    Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Add chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=Model=GenericPxieChassis')
        else:
            sys_def.chassis.add(chassis_number)

    # Define System Sync Modules (SSMs)
    if config.system_sync_modules_descriptors:
        interconnects = sys_def.interconnects
        ssm_list = []
        for descriptor in config.system_sync_modules_descriptors:
            if config.hardware_simulated:
                ssm = interconnects.add_sync_module(descriptor.resource_id, config.ssm_simulation_options)
            else:
                ssm = interconnects.add_sync_module(descriptor.resource_id, descriptor.options)
            # Append SSM object
            ssm_list.append(ssm)
        # Define connections between SSMs
        if config.ssm_connections:
            for connection in config.ssm_connections:
                connector_number = connection.ssm1_downstream_connector_number
                for ssm in ssm_list:
                    if ssm.chassis == connection.ssm1_chassis:
                        ssm1 = ssm
                    if ssm.chassis == connection.ssm2_chassis:
                        ssm2 = ssm
                # Implement each user-defined connection
                try:
                    # Set connection. SSMs have always one upstream port
                    ssm1.connectivity.systemsync_downstream[connector_number].set_connection(ssm2.connectivity.systemsync_upstream[0])
                except:
                    exit("Exception! Please check the valued defined for SyncModule resource ids, chassis numbers and connections")

    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_hvi_engines(sys_def, module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name, module in zip(module_dict.keys(), module_dict.values()):
        sys_def.engines.add(module.instrument.hvi.engines.main_engine, engine_name)


def define_hvi_actions(sys_def, module_dict, config):
    """ 
    define_hvi_actions(hvi, module_dict):

    hvi = HVI instance
    module_dict = dictionary containing modular instrument objects previously created
    config = object containing the application configuration
    
    This function defines a list of DAQ/AWG trigger actions for each module, 
    to be executed by the "action-execute" instructions within the HVI sequence.
    The number of actions in each engine's list depends on the intrument's number of channels.
    """
    # For each engine, add each HVI Actions to be executed to its own HVI Action Collection
    for engine_name, module in zip(module_dict.keys(), module_dict.values()):
        for ch_index in range(1, module.num_channels + 1):
            # Actions need to be added to the engine's action list so that they can be executed
            # Example: hvi.engines[i].actions.add(module_dict[i].hvi.actions.awg1_trigger, 'AWG1_trigger')
            if engine_name == config.dig_engine_name:
                action_name = config.daq_trigger_action_name+ str(ch_index) # arbitrary user-defined name
                instrument_action = "daq{}_trigger".format(ch_index) # name decided by instrument API
            else:
                action_name = config.awg_trigger_action_name+ str(ch_index) # arbitrary user-defined name
                instrument_action = "awg{}_trigger".format(ch_index) # name decided by instrument API
            action_id = getattr(module.instrument.hvi.actions, instrument_action)
            sys_def.engines[engine_name].actions.add(action_id, action_name)


def define_registers(sequencer, config):
    """
     Defines all registers for each HVI engine in the scope af the global sync sequence 
    """
    # Digitizer registers
    loops = sequencer.sync_sequence.scopes[config.dig_engine_name].registers.add(config.loops_name, kthvi.RegisterSize.SHORT)
    loops.initial_value = 0
    steps = sequencer.sync_sequence.scopes[config.dig_engine_name].registers.add(config.steps_name, kthvi.RegisterSize.SHORT)
    steps.initial_value = 0
    acq_delay = sequencer.sync_sequence.scopes[config.dig_engine_name].registers.add(config.acq_delay_name, kthvi.RegisterSize.SHORT)
    acq_delay.initial_value = 0
    hvi_done = sequencer.sync_sequence.scopes[config.dig_engine_name].registers.add(config.hvi_done_name, kthvi.RegisterSize.SHORT)
    hvi_done.initial_value = 0
    dig_counter = sequencer.sync_sequence.scopes[config.dig_engine_name].registers.add(config.dig_counter_name, kthvi.RegisterSize.SHORT)
    dig_counter.initial_value = 0
    # AWG registers
    awg_counter = sequencer.sync_sequence.scopes[config.awg_engine_name].registers.add(config.awg_counter_name, kthvi.RegisterSize.SHORT)
    awg_counter.initial_value = 0
    tau = sequencer.sync_sequence.scopes[config.awg_engine_name].registers.add(config.tau_name, kthvi.RegisterSize.SHORT)
    tau.initial_value = 0
    wfm_id = sequencer.sync_sequence.scopes[config.awg_engine_name].registers.add(config.wfm_id_name, kthvi.RegisterSize.SHORT)
    wfm_id.initial_value = 0
    pulse_length = sequencer.sync_sequence.scopes[config.awg_engine_name].registers.add(config.pulse_length_name, kthvi.RegisterSize.SHORT)
    pulse_length.initial_value = 0


def output_sys_def(sys_def, descriptors):
    """
    Outputs the details of each instrument and engine included in the system definition.
    """
    print("-------------------")
    print("System Definition")
    print("-------------------")
    # Output KS2201A HVI Core Version
    hvi_core_version = kthvi.SystemDefinition.hvi_core_version
    print("The PathWave Test Sync Executive API installed on your system has an HVI core version " + hvi_core_version.to_string() + "\n")
    # Output info about the HW instruments
    print("HW Instruments:")
    for descriptor in descriptors:
        engine = sys_def.engines[descriptor.engine_name]
        print("- Model: {} in chassis: {}, slot: {}, HVI Engine Name: {}, HVI core version: ".format(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, descriptor.engine_name) + 
        engine.software_version.to_string() + ", HVI Engine FPGA IP version: " + engine.firmware_version.to_string())
    # Output info about the Sync Modules
    if sys_def.interconnects:
        print("System Sync Modules:")
        for index in range(sys_def.interconnects.count):
            sync_module = sys_def.interconnects[index]
            len_upstream = len(sync_module.connectivity.systemsync_upstream)
            len_downstream = len(sync_module.connectivity.systemsync_downstream)
            print("- System Sync Module in chassis: {}, slot: {} with {} Upstream Ports and {} Downstream Ports".format(sync_module.chassis, sync_module.slot, len_upstream, len_downstream))


def program_dut_experiment(sequencer, module_dict, config):
    """
    This method programs the HVI sequence of this application.
    Different HVI statements are encapsulated as much as possible in separated SW methods to help users visualize
    the programmed HVI sequences.
    The programming example documentation on www.keysight.com contains an HVI diagram that graphically represents the programmed HVI sequence.
    """
    # Define registers within the scope of the outmost sync sequence
    define_registers(sequencer, config)
    
    # Define sync while condition
    loops = sequencer.sync_sequence.scopes[config.dig_engine_name].registers[config.loops_name]
    sync_while_condition = kthvi.Condition.register_comparison(loops, kthvi.ComparisonOperator.LESS_THAN, config.num_loops)
    # Add Sync While Statement
    sync_while = sequencer.sync_sequence.add_sync_while("Run Experiment Loops", 90, sync_while_condition)
    # Program experiment loops 
    program_experiment_loops(sync_while.sync_sequence, module_dict, config)

    # Add SMSB statement
    sync_block = sequencer.sync_sequence.add_sync_multi_sequence_block("Execution Completed", 260)
    # Program the SMSB to Complete the Execution
    program_execution_completed(sync_block, config)


def program_experiment_loops(sync_sequence, module_dict, config):
    """
    Programs the outmost Sync While statement that executes experiment loops
    """
    # Add 1st Sync Multi-Sequence Block to the Sync While sequence
    sync_block_1 = sync_sequence.add_sync_multi_sequence_block("Initialize registers", 260)
    # Program Sync Multi-Sequence Block: "Initialize registers"
    program_init_registers(sync_block_1, config)

    # Define sync while condition
    steps = sync_sequence.scopes[config.dig_engine_name].registers[config.steps_name]
    sync_while_condition = kthvi.Condition.register_comparison(steps, kthvi.ComparisonOperator.LESS_THAN, config.num_steps)
    # Add Sync While Statement
    sync_while_2 = sync_sequence.add_sync_while("Run Experiment Steps", 70, sync_while_condition)
    # Add and program 2nd Sync While: "Run Experiment Steps"
    program_experiment_steps(sync_while_2.sync_sequence, module_dict, config)

    # Add a 3rd Sync Multi-Sequence Block
    sync_block_3 = sync_sequence.add_sync_multi_sequence_block("Loop Delay", 260)
    # Program the 3rd SMSB "Loop Delay"
    program_loop_delay(sync_block_3, config)


def program_init_registers(sync_block, config):
    """
    Adds and programs the 1st Sync Multi-Sequence Block statement
    """
    ##### AWG Sequence #######
    initialize_awg_registers(sync_block.sequences[config.awg_engine_name], config)

    #### Digitizer Sequence ###
    initialize_dig_registers(sync_block.sequences[config.dig_engine_name], config)


def initialize_awg_registers(awg_sequence, config):
    """ 
    Initialize previously defined AWG registers 
    """
    # Initialize wfm_id = 0
    instruction = awg_sequence.add_instruction("wfm_id = 0", 10, awg_sequence.instruction_set.assign.id)
    instruction.set_parameter(awg_sequence.instruction_set.assign.destination.id, awg_sequence.scope.registers[config.wfm_id_name])
    instruction.set_parameter(awg_sequence.instruction_set.assign.source.id, 0)
    # Initialize tau = initial_tau
    instruction = awg_sequence.add_instruction("tau = initial_tau", 10, awg_sequence.instruction_set.assign.id)
    instruction.set_parameter(awg_sequence.instruction_set.assign.destination.id, awg_sequence.scope.registers[config.tau_name])
    instruction.set_parameter(awg_sequence.instruction_set.assign.source.id, config.initial_tau)
    # Initialize pulse_length = initial_pulse_length
    instruction = awg_sequence.add_instruction("pulse_length = initial_pulse_length", 10, awg_sequence.instruction_set.assign.id)
    instruction.set_parameter(awg_sequence.instruction_set.assign.destination.id, awg_sequence.scope.registers[config.pulse_length_name])
    instruction.set_parameter(awg_sequence.instruction_set.assign.source.id, config.initial_pulse_length)


def initialize_dig_registers(dig_sequence, config):
    """ 
    Initialize previously defined Digitizer registers 
    """
    # Initialize steps = 0
    instruction = dig_sequence.add_instruction("steps = 0", 10, dig_sequence.instruction_set.assign.id)
    instruction.set_parameter(dig_sequence.instruction_set.assign.destination.id, dig_sequence.scope.registers[config.steps_name])
    instruction.set_parameter(dig_sequence.instruction_set.assign.source.id, 0)
    # Initialize acq_delay = initial_acq_delay
    instruction = dig_sequence.add_instruction("acq_delay = initial_acq_delay", 10, dig_sequence.instruction_set.assign.id)
    instruction.set_parameter(dig_sequence.instruction_set.assign.destination.id, dig_sequence.scope.registers[config.acq_delay_name])
    instruction.set_parameter(dig_sequence.instruction_set.assign.source.id, config.initial_acq_delay)


def program_experiment_steps(sync_sequence, module_dict, config):
    """
    Programs the inner Sync While statement that executes experiment steps
    """
    # Previously defined registers
    steps = sync_sequence.scopes[config.dig_engine_name].registers[config.steps_name]
    wfm_id = sync_sequence.scopes[config.awg_engine_name].registers[config.wfm_id_name]
    
    # Add sync register sharing
    bits_to_share = 3
    sync_sequence.add_sync_register_sharing("Share steps->wfm_id", 260, steps, wfm_id, bits_to_share)

    # Add a Sync Multi-Sequence Block
    sync_block_2 = sync_sequence.add_sync_multi_sequence_block("Generate waveforms", 10)
    # Program the SMSB to Generate Waveforms
    program_generate_waveforms(sync_block_2, module_dict, config)


def program_generate_waveforms(sync_block, module_dict, config):
    """
    Programs the SMSB that generates waveforms
    """
    ##### AWG Sequence #######
    awg_module = module_dict[config.awg_engine_name].instrument
    awg_sequence = sync_block.sequences[config.awg_engine_name]
    awg_experiment(awg_sequence, awg_module, config)

    ### Digitizer Sequence ###
    dig_module = module_dict[config.dig_engine_name]
    dig_sequence = sync_block.sequences[config.dig_engine_name]
    dig_experiment(dig_sequence, dig_module, config)


def awg_experiment(awg_sequence, awg_module, config):
    """
    Programs the local awg sequence containing the core experiment routine
    """
    # Previously defined registers
    wfm_id = awg_sequence.scope.registers[config.wfm_id_name]
    tau = awg_sequence.scope.registers[config.tau_name]
    pulse_length = awg_sequence.scope.registers[config.pulse_length_name]
    awg_counter = awg_sequence.scope.registers[config.awg_counter_name]

    # Previously defined actions to be executed within the experiment
    awg_trigger_action_12 = [
        awg_sequence.engine.actions[config.awg_trigger_action_name + str(1)], 
        awg_sequence.engine.actions[config.awg_trigger_action_name + str(2)]]
    awg_trigger_action_34 = [
        awg_sequence.engine.actions[config.awg_trigger_action_name + str(3)], 
        awg_sequence.engine.actions[config.awg_trigger_action_name + str(4)]]

    # Define If condition and parameters
    if_condition = kthvi.Condition.register_comparison(wfm_id, kthvi.ComparisonOperator.GREATER_THAN_OR_EQUAL_TO, config.num_wfms)
    enable_ifbranches_time_matching = True
    # Add If statement
    if_statement = awg_sequence.add_if("Check wfm_id", 70, if_condition, enable_ifbranches_time_matching)
    if_branch_seq = if_statement.if_branch.sequence
    # Reset wfm_id = 0 within the IF sequence
    instruction = if_branch_seq.add_instruction("wfm_id = 0", 30, awg_sequence.instruction_set.assign.id)
    instruction.set_parameter(awg_sequence.instruction_set.assign.destination.id, wfm_id)
    instruction.set_parameter(awg_sequence.instruction_set.assign.source.id, 0)
    
    # Queue waveform to CH1, CH2
    # NOTE: the next instructions needs a start delay of at least 60 ns to make sure wfm_id is updated
    for awg_ch in range(1, 3):
        instrLabel = "QueueWaveform(CH" + str(awg_ch) + " , wfm_id)"
        instruction0 = awg_sequence.add_instruction(instrLabel, config.queue_wfm_latency, awg_module.hvi.instruction_set.queue_waveform.id)
        #Set every parameter of AWGqueueWaveform(awg_ch, waveformNumber, triggerMode, startDelay, cycles, prescaler);
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.waveform_number.id, wfm_id)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.channel.id, awg_ch)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.trigger_mode.id, config.awg_trigger_mode)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.start_delay.id, config.start_delay)
        if config.T2_flag:
            instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.cycles.id, config.wfm_cycles)
        else:
            instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.cycles.id, 1)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.prescaler.id, config.awg_prescaler)
    
    # Queue Readout rising edge waveform to CH3, CH4
    for awg_ch in range(3, 5):
        instrLabel = "Queue ROwaveRise CH" + str(awg_ch)
        instruction0 = awg_sequence.add_instruction(instrLabel, config.queue_wfm_latency, awg_module.hvi.instruction_set.queue_waveform.id)
        #Set every parameter of AWGqueueWaveform(awg_ch, waveformNumber, triggerMode, startDelay, cycles, prescaler);
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.waveform_number.id, config.rorise_id)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.channel.id, awg_ch)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.trigger_mode.id, config.awg_trigger_mode)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.start_delay.id, config.start_delay)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.cycles.id, 1)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.prescaler.id, config.awg_prescaler)
    
    # Queue Readout falling edge waveform to CH3, CH4
    for awg_ch in range(3, 5):
        instrLabel = "Queue ROwaveFall CH" + str(awg_ch)
        instruction0 = awg_sequence.add_instruction(instrLabel, config.queue_wfm_latency, awg_module.hvi.instruction_set.queue_waveform.id)
        #Set every parameter of AWGqueueWaveform(awg_ch, waveformNumber, triggerMode, startDelay, cycles, prescaler);
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.waveform_number.id, config.rofall_id)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.channel.id, awg_ch)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.trigger_mode.id, config.awg_trigger_mode)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.start_delay.id, config.start_delay)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.cycles.id, 1)
        instruction0.set_parameter(awg_module.hvi.instruction_set.queue_waveform.prescaler.id, config.awg_prescaler)
    
    # AWG trigger CH1, CH2 - Generates first pulse
    inst_awg_trigger = awg_sequence.add_instruction("AwgTrigger(CH1, CH2)", config.awg_trigger_latency, awg_sequence.instruction_set.action_execute.id)
    inst_awg_trigger.set_parameter(awg_sequence.instruction_set.action_execute.action.id, awg_trigger_action_12)
    
    # WaitTime: tau
    awg_sequence.add_wait_time("WaitTime: tau", 30, tau)
    
    # Differentiate between T1 and T2 experiments
    if config.T2_flag:
        # AWG trigger CH1, CH2 - Generates second pulse
        inst_awg_trigger = awg_sequence.add_instruction("2nd AwgTrigger(CH1, CH2)", 10, awg_sequence.instruction_set.action_execute.id)
        inst_awg_trigger.set_parameter(awg_sequence.instruction_set.action_execute.action.id, awg_trigger_action_12)
    
    # AWG trigger CH3, CH4 - Generates readout pulse rising edge
    inst_awg_trigger = awg_sequence.add_instruction("AwgTrigger(CH3, CH4)", config.ro_delay, awg_sequence.instruction_set.action_execute.id)
    inst_awg_trigger.set_parameter(awg_sequence.instruction_set.action_execute.action.id, awg_trigger_action_34)
    
    # WaitTime: pulse_length - Holds on the readout pulse amplitude
    awg_sequence.add_wait_time("WaitTime: pulse_length", 30, pulse_length)
    
    # AWG trigger CH3, CH4 - Generates readout pulse falling edge
    inst_awg_trigger = awg_sequence.add_instruction("2nd AwgTrigger(CH3, CH4)", 10, awg_sequence.instruction_set.action_execute.id)
    inst_awg_trigger.set_parameter(awg_sequence.instruction_set.action_execute.action.id, awg_trigger_action_34)
    
    # tau += tau_step
    instruction = awg_sequence.add_instruction("tau += tau_step", 10, awg_sequence.instruction_set.add.id)
    instruction.set_parameter(awg_sequence.instruction_set.add.destination.id, tau)
    instruction.set_parameter(awg_sequence.instruction_set.add.left_operand.id, tau)
    instruction.set_parameter(awg_sequence.instruction_set.add.right_operand.id, config.tau_step)
    
    # pulse_length += delta_length
    instruction = awg_sequence.add_instruction("pulse_length += delta_length", 10, awg_sequence.instruction_set.add.id)
    instruction.set_parameter(awg_sequence.instruction_set.add.destination.id, pulse_length)
    instruction.set_parameter(awg_sequence.instruction_set.add.left_operand.id, pulse_length)
    instruction.set_parameter(awg_sequence.instruction_set.add.right_operand.id, config.delta_length)
    
    # awg_counter++
    instruction = awg_sequence.add_instruction("awg_counter++", 10, awg_sequence.instruction_set.add.id)
    instruction.set_parameter(awg_sequence.instruction_set.add.destination.id, awg_counter)
    instruction.set_parameter(awg_sequence.instruction_set.add.left_operand.id, awg_counter)
    instruction.set_parameter(awg_sequence.instruction_set.add.right_operand.id, 1)
    
    # Step delay
    awg_sequence.add_delay("Step Delay", config.step_delay)


def dig_experiment(dig_sequence, dig_module, config):
    """
    Programs the local digitizer sequence containing the core experiment routine
    """
    # Previously defined registers
    steps = dig_sequence.scope.registers[config.steps_name]
    acq_delay = dig_sequence.scope.registers[config.acq_delay_name]
    dig_counter = dig_sequence.scope.registers[config.dig_counter_name]

    # Define DAQ trigger action from previously defined digitizer actions
    daq_trigger_all = []
    for ch_index in range(1, dig_module.num_channels+1):
        daq_trigger_all.append(dig_sequence.engine.actions[ch_index-1])

    # WaitTime: acq_delay
    dig_sequence.add_wait_time("WaitTime: acq_delay", 30, acq_delay)

    # DAQ Trigger action execute
    # Action-execute instruction
    inst_daq_trigger = dig_sequence.add_instruction("DaqTrigger(CH1, CH2, CH3, CH4)", 10, dig_sequence.instruction_set.action_execute.id)
    inst_daq_trigger.set_parameter(dig_sequence.instruction_set.action_execute.action.id, daq_trigger_all)
    
    # Increment steps
    instruction = dig_sequence.add_instruction("steps++", 10, dig_sequence.instruction_set.add.id)
    instruction.set_parameter(dig_sequence.instruction_set.add.destination.id, steps)
    instruction.set_parameter(dig_sequence.instruction_set.add.left_operand.id, steps)
    instruction.set_parameter(dig_sequence.instruction_set.add.right_operand.id, 1)
    
    # Increment dig_counter
    instruction = dig_sequence.add_instruction("dig_counter++", 10, dig_sequence.instruction_set.add.id)
    instruction.set_parameter(dig_sequence.instruction_set.add.destination.id, dig_counter)
    instruction.set_parameter(dig_sequence.instruction_set.add.left_operand.id, dig_counter)
    instruction.set_parameter(dig_sequence.instruction_set.add.right_operand.id, 1)
    
    # acq_delay += tau_step
    instruction = dig_sequence.add_instruction("acq_delay += tau_step", 10, dig_sequence.instruction_set.add.id)
    instruction.set_parameter(dig_sequence.instruction_set.add.destination.id, acq_delay)
    instruction.set_parameter(dig_sequence.instruction_set.add.left_operand.id, acq_delay)
    instruction.set_parameter(dig_sequence.instruction_set.add.right_operand.id, config.tau_step)

    return


def program_loop_delay(sync_block, config):
    """
    Programs the loop delay statements in the digitizer local sequence
    """
    # Previously defined registers
    loops = sync_block.sequences[config.dig_engine_name].scope.registers[config.loops_name]
    
    #### Digitizer Sequence ####
    dig_sequence = sync_block.sequences[config.dig_engine_name]
    # Increment loops
    instruction = dig_sequence.add_instruction("loops++", 10, dig_sequence.instruction_set.add.id)
    instruction.set_parameter(dig_sequence.instruction_set.add.destination.id, loops)
    instruction.set_parameter(dig_sequence.instruction_set.add.left_operand.id, loops)
    instruction.set_parameter(dig_sequence.instruction_set.add.right_operand.id, 1)
    
    # Loop delay
    dig_sequence.add_delay("Loop Delay", config.loop_delay)


def program_execution_completed(sync_block, config):
    """
    Adds and programs the 4th Sync Multi-Sequence Block statement
    """
    # Previously defined registers
    reg_hvi_done = sync_block.sequences[config.dig_engine_name].scope.registers[config.hvi_done_name]
    
    # Set hvi_done = 1 in digitizer module
    dig_sequence = sync_block.sequences[config.dig_engine_name]
    instruction = dig_sequence.add_instruction("hvi_done = 1", 10, dig_sequence.instruction_set.assign.id)
    instruction.set_parameter(dig_sequence.instruction_set.assign.destination.id, reg_hvi_done)
    instruction.set_parameter(dig_sequence.instruction_set.assign.source.id, 1)


def export_hvi_seqeunces(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    print("Programmed HVI sequences exported to file")
    
    # Write file
    text_file = open("TextOutput/HviDutExperiment.txt", "w")
    text_file.write(output)
    text_file.close()


def interact_with_hvi(hvi, module_dict, config):
    """
    This SW method compiles HVI, loads it to HW and executes it
    """
    # Access previously defined resources
    digitizer_module = module_dict[config.dig_engine_name].instrument
    # Dig registers
    dig_registers = hvi.sync_sequence.scopes[config.dig_engine_name].registers # digitizer registers collection
    steps = dig_registers[config.steps_name]
    loops = dig_registers[config.loops_name]
    acq_delay = dig_registers[config.acq_delay_name]
    dig_counter = dig_registers[config.dig_counter_name]
    hvi_done = dig_registers[config.hvi_done_name]
    # AWG registers
    awg_registers = hvi.sync_sequence.scopes[config.awg_engine_name].registers
    wfm_id = awg_registers[config.wfm_id_name]
    tau = awg_registers[config.tau_name]
    pulse_length = awg_registers[config.pulse_length_name]
    awg_counter = awg_registers[config.awg_counter_name]

    # User controlled loop
    attempts = 0
    max_attempts = 50
    hvi_done_read = hvi_done.read()
    while (hvi_done_read == 0) and (attempts<max_attempts):
        time.sleep(1) #waits for 1s
        hvi_done_read = hvi_done.read()
        attempts += 1
    if attempts>=max_attempts:
        print("Timeout.. Quitting the application execution...")
    print()
    print("------ Final Register Values ------")
    print("Experiment steps: {}".format(steps.read()))
    print("Experiment loops: {}".format(loops.read()))
    print("Final awg_counter: {}".format(awg_counter.read()))
    print("Final dig_counter: {}".format(dig_counter.read()))
    print("Final wfm_id: {}".format(wfm_id.read()))
    print("Final tau: {}".format(tau.read()))
    print("Final acq_delay: {}".format(acq_delay.read()))
    print("Final pulse_length: {}".format(pulse_length.read()))
    print()
    
    if hvi_done.read() == 1:
        print("HVI Execution Completed Successfully!")

        # DAQstop: stops any previous acquisition
        digitizer_module.DAQstopMultiple(config.dig_mask)
        
        # Digitizer acquisition
        timeout = 500 #[ms]
        meas_points = config.acquisition_points
        measured_data1 = numpy.zeros(meas_points)
        measured_data1 = digitizer_module.DAQread(1, meas_points, timeout)
        measured_data2 = numpy.zeros(meas_points)
        measured_data2 = digitizer_module.DAQread(3, meas_points, timeout)

        # Plot acquisition data
        plt.clf()
        plt.plot(measured_data1, 'b-', measured_data2, 'r-')
        plt.show()
    else:
        print("HVI Execution Failed!")

    return hvi_done_read

#################################################################
# 3rd Level: Classes and Functions to Use SD1/M3xxxA Instruments
#################################################################

class ModuleDescriptor:
    "Descriptor for module objects"
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class SystemSyncModuleDescriptor:
    """
    Descriptor for System Sync Modules
    """
    def __init__(self, resource_id, options):
        self.resource_id = resource_id
        self.options = options


class SystemSyncModuleConnection:
    """
    Describes the connection between two SyncModule objects ssm1 and ssm2. SSMs have always one upstream port:
    ssm1.connectivity.systemsync_downstream[ssm1_downstream_connector_number].connection = ssm2.connectivity.systemsync_upstream[0]
    """
    def __init__(self, ssm1_chassis, ssm1_downstream_connector_number, ssm2_chassis):
        self.ssm1_chassis = ssm1_chassis
        self.ssm1_downstream_connector_number = ssm1_downstream_connector_number
        self.ssm2_chassis = ssm2_chassis


class Module:
    "Class defining a modular instrument object and its properties"
    def __init__(self, instrument_object, num_channels):
        self.instrument = instrument_object
        self.num_channels = num_channels


def configure_awg(config, awg_module, num_channels = 4):
    """ 
    configure_awg(awg_module, num_channels) configures an awg_module to output arbitrary waveforms.
    Waveforms are going to be queued and triggered from the HVI sequence. 
    """
    # This application assumes num_channels = 4
    if(num_channels != 4):
        raise Exception("This example assumes an AWG with 4 channels is used. Exiting... ")

    # Load waveform to AWG memory
    awg_module.waveformFlush() #memory flush
    wave = keysightSD1.SD_Wave()
    wave.newFromFile("Waveforms/Piwave.csv")
    awg_module.waveformLoad(wave, 0)
    wave.newFromFile("Waveforms/Exponential.csv")
    awg_module.waveformLoad(wave, 1)
    wave.newFromFile("Waveforms/SquarePulse.csv")
    awg_module.waveformLoad(wave, 2)
    wave.newFromFile("Waveforms/TriangularPulse.csv")
    awg_module.waveformLoad(wave, 3)
    wave.newFromFile("Waveforms/ROwave.csv")
    awg_module.waveformLoad(wave, 4)
    wave.newFromFile("Waveforms/Pi2wave.csv")
    awg_module.waveformLoad(wave, 5)
    wave.newFromFile("Waveforms/ROwaveRise.csv")
    awg_module.waveformLoad(wave, config.rorise_id)
    wave.newFromFile("Waveforms/ROwaveFall.csv")
    awg_module.waveformLoad(wave, config.rofall_id)
    if config.num_wfms > 6:
        raise Exception(
        "The num_wfm parameter value is exceeding the number of waveforms loaded to the AWG RAM. "
        "Please modify the code to load more waveform or use a lower num_wfm value. Exiting... ")

    for awg_ch in range(1, num_channels+1):
        # AWG queue flush 
        awg_module.AWGstop(awg_ch)
        awg_module.AWGflush(awg_ch)
        
        # AWG configuration
        awg_module.AWGqueueConfig(awg_ch, config.queue_mode)
        awg_module.AWGqueueSyncMode(awg_ch, config.sync_mode)

        # Set AWG channel properties
        awg_module.channelWaveShape(awg_ch, config.awg_mode)
        awg_module.modulationAmplitudeConfig(awg_ch, keysightSD1.SD_ModulationTypes.AOU_MOD_AM, config.amplitude)
        awg_module.channelAmplitude(awg_ch, 0)
        awg_module.channelOffset(awg_ch, config.offset)
        awg_module.channelFrequency(awg_ch, config.carrier_frequency)
        if awg_ch == 1 or awg_ch == 3:
            awg_module.channelPhase(awg_ch, 0)
        else:
            awg_module.channelPhase(awg_ch, 90)
        
        # Enable AWGfreezeOnStop to hold on the voltage value of the last wfm sample
        awg_module.AWGfreezeOnStopEnable(awg_ch, 1)
        
    # Start AWG
    awg_module.AWGstartMultiple(config.all_ch_mask) # AWG channels start and wait for trigger

    return


def configure_digitizer(config, digitizer_module, num_channels = 4):
    """ 
    This method configures multiple DAQ channels for cyclic acquisitions
    """
    # Input settings
    prescaler = config.dig_prescaler 
    fullscale = config.fullscale
    num_cycles = config.num_cycles
    points_per_cycle = config.acquisition_points_per_cycle
    acquisition_delay = config.acquisition_delay
    trigger_mode = config.dig_trigger_mode
    
    # DAQstop: stops any previous acquisition
    digitizer_module.DAQstopMultiple(config.dig_mask)

    # Configure DAQ channels
    for n_DAQ in range (1, num_channels+1):
        # Resets the DAQ
        digitizer_module.DAQflush(n_DAQ)

        # Digitizer channel configuration
        digitizer_module.channelInputConfig(n_DAQ, fullscale, keysightSD1.AIN_Impedance.AIN_IMPEDANCE_50, keysightSD1.AIN_Coupling.AIN_COUPLING_DC)
        digitizer_module.channelPrescalerConfig(n_DAQ, prescaler)

        # DAQ acquisitions configuration
        digitizer_module.DAQconfig(n_DAQ, points_per_cycle, num_cycles, acquisition_delay, trigger_mode)
    
    # DAQstart: digitizer channels get strted and are waiting for the DAQ trigger commands that trigger the acquisition of each cycle of data points
    digitizer_module.DAQstartMultiple(config.dig_mask)
    
    return points_per_cycle


def open_modules(config):
    """ 
    Opens and creates all the necessary instrument objects.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Guide for options to open the instrument objects
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Open SD1 instrument objects
    for descriptor in config.module_descriptors:
        if descriptor.model_number == "M3102A":
            instr_obj = keysightSD1.SD_AIN()
        else:
            instr_obj = keysightSD1.SD_AOU()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {} - {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        module_dict[descriptor.engine_name] = Module(instr_obj, num_channels)
        num_modules += 1
    
    return module_dict


if __name__ == '__main__':
    run_experiment()